/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2012 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.opt.oo;

import java.lang.annotation.Annotation;
import java.lang.reflect.Modifier;

import org.oasisopen.sca.annotation.Callback;
import org.oasisopen.sca.annotation.OneWay;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.ComponentInterface;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.Utils;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.desc.MembraneDesc;
import org.objectweb.fractal.juliac.proxy.AbstractProxyClassGenerator;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedClassHelper;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;
import org.objectweb.fractal.juliac.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.visit.ClassSourceCodeVisitor;
import org.osoa.sca.annotations.Conversational;
import org.osoa.sca.annotations.EndsConversation;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.oasis.ServiceReferenceImpl;
import org.ow2.frascati.tinfi.osoa.ConversationImpl;
import org.ow2.frascati.tinfi.reflect.Util;

/**
 * This class generates the source code of SCA service reference
 * implementations.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 */
public class ServiceReferenceClassGenerator
extends AbstractProxyClassGenerator {
    
    public ServiceReferenceClassGenerator() {}
    
    public ServiceReferenceClassGenerator(
        Juliac jc,
		InterfaceType it, MembraneDesc<?> membraneDesc, boolean mergeable ) {
        
        super(jc,it,membraneDesc,mergeable);
    }
    
    /**
     * Return <code>true</code> if some code must be generated by this
     * generator.
     */
    public boolean match() {
        return true;
    }

    /**
     * Return the name of the suffix which is appended to generated class names.
     */
    public String getClassNameSuffix() {
        return "FcSR";
    }

    @Override
    public void setMergeable( boolean mergeable ) {
        if(mergeable) {
            String msg = "This generator can not generate mergeable code";
            throw new IllegalArgumentException(msg);
        }
    }

    @Override
    public String getSuperClassName() {

        StringBuffer sb = new StringBuffer();
        sb.append(ServiceReferenceImpl.class.getName());
        sb.append('<');
    	
        /*
         * Add type parameters if any.
         */
        String signature = it.getFcItfSignature();
        sb.append(signature);
        UnifiedClass proxycl = jc.create(signature);
        String[] tpnames = proxycl.getTypeParameterNames();
        if( tpnames.length != 0 ) {
            String s = Utils.getTypeParameterNamesSignature(tpnames);
            sb.append(s);
        }
        
        sb.append('>');
        return sb.toString();
    }

    @Override
    public String[] getImplementedInterfaceNames() {

        /*
         * Create the signature of the proxy interface with generic parameters
         * if any.
         */
    	String signature = it.getFcItfSignature();
        UnifiedClass proxycl = jc.create(signature);
        String[] tpnames = proxycl.getTypeParameterNames();
        if( tpnames.length != 0 ) {
            String s = Utils.getTypeParameterNamesSignature(tpnames);
            signature += s;
        }
        
        return new String[]{signature};
    }
    
    @Override
    public void generateConstructors(
        ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
        
        super.generateConstructors(cv,proxycl);
        
        /*
         * Create the signature of the proxy interface with generic parameters
         * if any.
         */
    	String signature = it.getFcItfSignature();
        String[] tpnames = proxycl.getTypeParameterNames();
        if( tpnames.length != 0 ) {
            String s = Utils.getTypeParameterNamesSignature(tpnames);
            signature += s;
        }        
        
        // public constructor(Class<B>,B)
        String[] parameters =
            new String[]{
        		"Class<"+signature+"> businessInterface",
        		signature+" service" };
        BlockSourceCodeVisitor mv =
            cv.visitConstructor(Modifier.PUBLIC,null,parameters,null);
        mv.visitIns("super(businessInterface,service)");
        mv.visitEnd();
    }
    
    /**
     * @since 1.4.1
     */
    @Override
    public void generateMethods( ClassSourceCodeVisitor cv, UnifiedClass proxycl ) {
    	/*
    	 * Override the getService method inherited from ServiceReferenceImpl.
    	 * The idea is that in generated subclasses of ServiceReferenceImpl we
    	 * can directly return "this" for getService. Whereas when getService is
    	 * invoked on an instance of ServiceReferenceImpl, getService should
    	 * return an instance which implements both the ServiceReference
    	 * interface and the business interface. This is the case of generated
    	 * subclass instances.
    	 */
    	String signature = it.getFcItfSignature();
    	BlockSourceCodeVisitor mv =
    		cv.visitMethod(
				Modifier.PUBLIC, null, signature, "getService", null, null );
    	mv.visitIns("return this");
    	mv.visitEnd();
    }
    
    @Override
    public void generateProxyMethodBodyBeforeCode(
		BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        super.generateProxyMethodBodyBeforeCode(mv,proxycl,proxym);
    
        conversationalItf = proxycl.getAnnotation(Conversational.class);
        oneWayMeth =
        	UnifiedClassHelper.getAnnotation(
    			proxym,
    			OneWay.class.getName(),
    			"org.osoa.sca.annotations.OneWay");
        
        callbackClass = Void.class;   // callback.value() defaults to Void
        Annotation callback =
        	UnifiedClassHelper.getAnnotation(
    			proxycl,
    			Callback.class.getName(),
    			"org.osoa.sca.annotations.Callback");
        if( callback != null ) {
            callbackClass = Util.getAnnotationParamValue(callback,"value");
        }

        // OneWay method
        if( oneWayMeth != null ) {
        	UnifiedClass rtype = proxym.getReturnType();
        	if( ! rtype.getName().equals("void") ) {
        		String msg =
        			"@OneWay annotated method "+proxym.toString()+" in "+
        			proxycl.getName()+" should return void";
        		throw new TinfiRuntimeException(msg);
        	}
            mv.visitln("    new Thread() {");
            mv.visitln("      public void run() {");
        }

        // Method defined in a conversational interface
        if( conversationalItf != null ) {
            mv.visitln("        "+ConversationImpl.class.getName()+" conv = _getConversation();");
            mv.visitln("        _pushConversation(conv);");
        }
        
        // Method defined in an interface with a callback
        if( callbackClass != Void.class ) {
            
            InterfaceType cbit =
                new BasicInterfaceType(
                    "callback",callbackClass.getName(),false,false,false);
            ClassGeneratorItf cg =
                new CallBackInterfaceClassGenerator(jc,cbit,null,mergeable);
            String cbitfclname = cg.getTargetClassName();
            
            cg = new ServiceReferenceClassGenerator(jc,cbit,null,mergeable);
            String cbcrclname = cg.getTargetClassName();
            
            mv.visit  ("        "+ComponentInterface.class.getName()+" ci = (");
            mv.visitln(ComponentInterface.class.getName()+") service;");
            mv.visitln("        "+Component.class.getName()+" owner = ci.getFcItfOwner();");
            mv.visit  ("        "+InterfaceType.class.getName()+" it = ");
            mv.visit  (Utils.javaify(cbit).toString());
            mv.visitln(";");
            mv.visitln("        "+cbitfclname+" cbci = new "+cbitfclname+"(owner,\"callback\",it,false,null);");
            
            mv.visit  ("        "+ServiceReferenceImpl.class.getName()+"<"+callbackClass.getName()+"> cb = ");
            mv.visit  ("new "+cbcrclname+"(");
            mv.visitln(callbackClass.getName()+".class,cbci);");
            mv.visitln("        _setCallback(cb);");
            mv.visitln("        _pushCallback(cb);");
        }
    }
    private Conversational conversationalItf;
    private Annotation oneWayMeth;
    private Class<?> callbackClass;
    
    @Override
    public void generateProxyMethodBodyAfterCode(
		BlockSourceCodeVisitor mv, UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        super.generateProxyMethodBodyAfterCode(mv,proxycl,proxym);
        
        EndsConversation end = proxym.getAnnotation(EndsConversation.class);
        if( end != null ) {
            if( conversationalItf == null ) {
            	/*
            	 * For cases where @EndsConversation appears in non
            	 * @Conversation annotated interfaces. That can be the case for
            	 * example for callback interfaces associated with conversation
            	 * scoped components.
            	 */
                mv.visitln("    "+ConversationImpl.class.getName()+" conv = _peekConversation();");
            }
            mv.visitln("    conv.end();");
        }
        if( conversationalItf != null ) {
            mv.visitln("    conv.access();");
            mv.visitln("    _popConversation();");
        }
        if( callbackClass != Void.class ) {
            mv.visitln("        _popCallback();");
        }
        if( oneWayMeth != null ) {
            mv.visitln("      }");
            mv.visitln("    }.start();");
        }
    }

    public String getDelegatingInstance(
        UnifiedClass proxycl, UnifiedMethod proxym ) {
        
        return "service";
    }
}
